/*
 * pclt.c - Process Image client
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <errno.h>

#include "xmalloc.h"
#include "buffer.h"
#include "list.h"
#include "tree.h"
#include "str.h"
#include "sock-unix.h"
#include "prioq.h"
#include "io.h"
#include "timeout.h"
#include "par.h"
#include "pimage.h"
#include "log.h"
#include "pi-proto.h"
#include "pclt.h"

/*#define DEBUG_PI_CLIENT 1*/

#ifdef DEBUG_PI_CLIENT
#define debug(fmt, arg...) debug(fmt, ##arg)
#else
#define	debug(fmt, arg...) do {} while (0)
#endif

#define PEER_TIMEOUT_MSEC	500

static struct io_context *io = NULL;
static slist_t client_list = NULL;

struct client_session {
	pi_side_t		side;			/* master or slave? */
	int			sfd;			/* socket fd */
	buffer_t		wb;			/* write buffer */
	buffer_t		rb;			/* read buffer */
	int			state;			/* state of session */
	size_t			last_offset;		/* offset used to track partial read/write */
	size_t			peer_inputs_sz;		/* peer's inputs image size */
	size_t			peer_outputs_sz;	/* peer's outputs image size */
	size_t			inputs_sz;		/* actual inputs image size */
	size_t			outputs_sz;		/* actual outputs image size */
	struct par_image_struct	*pi;			/* parameters image */
	struct timeout		*tout;			/* read timeout */
	int			status;			/* last operation status */
	int			response_timeout;	/* msecs to wait for response or confirmation */
	char			sockfile[256];		/* socket file to connect to */
};

struct pclt {
	struct client_session	s;
};

static void monitor_state();
static void client_session_disable(struct client_session *s, int reason);

static void client_send(struct client_session *s)
{
	int res;

	res = buffer_write_fd(s->sfd, s->wb);

	if (res == -1 && (errno != EINTR &&
			errno != EAGAIN && errno != EWOULDBLOCK)) {
		warn("client_send: write()=%d sock=%d: %s",
				res, s->sfd, strerror(errno));
	}

	if (buffer_size(s->wb) != 0)
		io_change(io, s->sfd, EVENT_READ|EVENT_WRITE);
}

static void client_connect(struct client_session *s)
{
	/* Send HELLO to server */
	debug("Sending HELLO!");
	buffer_put_u16(s->wb, 2);
	buffer_put_u8(s->wb, CMD_CONNECT);
	buffer_put_u8(s->wb, 0);
	client_send(s);
	timeout_add(s->tout, s->response_timeout);
}

static void client_inputs_read(struct client_session *s, size_t len, size_t offset)
{
	debug("Sending INPUTS_READ %lu at %lu", len, offset);
	buffer_put_u16(s->wb, 6);
	buffer_put_u8(s->wb, CMD_INPUTS_READ);
	buffer_put_u8(s->wb, STATUS_OK);
	buffer_put_u16(s->wb, len);
	buffer_put_u16(s->wb, offset);
	client_send(s);
	timeout_add(s->tout, s->response_timeout);
}

static void client_outputs_read(struct client_session *s, size_t len, size_t offset)
{
	debug("Sending OUTPUTS_READ %lu at %lu", len, offset);
	buffer_put_u16(s->wb, 6);
	buffer_put_u8(s->wb, CMD_OUTPUTS_READ);
	buffer_put_u8(s->wb, STATUS_OK);
	buffer_put_u16(s->wb, len);
	buffer_put_u16(s->wb, offset);
	client_send(s);
	timeout_add(s->tout, s->response_timeout);
}

static void client_obtain_config(struct client_session *s)
{
	debug("Sending OBTAIN_CONFIG");
	buffer_put_u16(s->wb, 2);
	buffer_put_u8(s->wb, CMD_OBTAIN_CONFIG);
	buffer_put_u8(s->wb, STATUS_OK);
	client_send(s);
	timeout_add(s->tout, s->response_timeout);
}

static void client_prepare_read(struct client_session *s)
{
	debug("Sending PREPARE_READ");
	buffer_put_u16(s->wb, 2);
	buffer_put_u8(s->wb, CMD_PREPARE_READ);
	buffer_put_u8(s->wb, STATUS_OK);
	client_send(s);
	timeout_add(s->tout, s->response_timeout);
}

static void client_prepare_write(struct client_session *s)
{
	debug("Sending PREPARE_WRITE");
	buffer_put_u16(s->wb, 2);
	buffer_put_u8(s->wb, CMD_PREPARE_WRITE);
	buffer_put_u8(s->wb, STATUS_OK);
	client_send(s);
	timeout_add(s->tout, s->response_timeout);
}

static void client_end_read(struct client_session *s)
{
	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		debug("fd=%d Sending END_READ", s->sfd);
		buffer_put_u16(s->wb, 2);
		buffer_put_u8(s->wb, CMD_END_READ);
		buffer_put_u8(s->wb, STATUS_OK);
		client_send(s);
		timeout_add(s->tout, s->response_timeout);
		s->state = CLIENT_END_READ_WAIT;
		debug("fd=%d CONNECTED->END_READ_WAIT", s->sfd);
	} else {
		warn("client_end_read: sfd=%d", s->sfd);
	}
}

static void client_end_write(struct client_session *s)
{
	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		debug("fd=%d Sending END_WRITE", s->sfd);
		buffer_put_u16(s->wb, 2);
		buffer_put_u8(s->wb, CMD_END_WRITE);
		buffer_put_u8(s->wb, STATUS_OK);
		client_send(s);
		timeout_add(s->tout, s->response_timeout);
		s->state = CLIENT_END_WRITE_WAIT;
		debug("fd=%d CONNECTED->END_WRITE_WAIT", s->sfd);
	} else {
		warn("client_end_write: sfd=%d", s->sfd);
	}
}

/* notify host that we want inputs */
static void client_prepare_inputs(struct client_session *s)
{
	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		client_prepare_read(s);
		s->state = CLIENT_PREPARE_READ_WAIT;
		debug("fd=%d CONNECTED->PREPARE_READ_WAIT", s->sfd);
	} else {
		warn("client_prepare_inputs: sfd=%d", s->sfd);
	}
}

/* notify host that we want outputs */
static void client_prepare_outputs(struct client_session *s)
{
	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		client_prepare_write(s);
		s->state = CLIENT_PREPARE_WRITE_WAIT;
		debug("fd=%d CONNECTED->PREPARE_WRITE_WAIT", s->sfd);
	} else {
		warn("client_prepare_outputs: sfd=%d", s->sfd);
	}
}

/* read inputs from server */
static void client_get_inputs(struct client_session *s)
{
	size_t len;

	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		if (s->inputs_sz) {
			s->last_offset = 0;
			len = s->inputs_sz > PEER_MAX_MSG-6 ?
					PEER_MAX_MSG-6 : s->inputs_sz;
			client_inputs_read(s, len, 0);
			s->last_offset += len;
			s->state = CLIENT_INPUTS_WAIT;
			debug("fd=%d CONNECTED->INPUTS_WAIT", s->sfd);
		} else {
			debug("fd=%d no inputs to read", s->sfd);
		}
	} else {
		warn("client_get_inputs(): sfd=%d", s->sfd);
	}
}

/* read outputs from server */
static void client_get_outputs(struct client_session *s)
{
	size_t len;

	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		if (s->outputs_sz) {
			s->last_offset = 0;
			len = s->outputs_sz > PEER_MAX_MSG-6 ?
					PEER_MAX_MSG-6 : s->outputs_sz;
			client_outputs_read(s, len, 0);
			s->last_offset += len;
			s->state = CLIENT_OUTPUTS_WAIT;
			debug("fd=%d CONNECTED->OUTPUTS_WAIT", s->sfd);
		} else {
			debug("fd=%d no outputs to read", s->sfd);
		}
	} else {
		warn("client_get_outputs(): sfd=%d", s->sfd);
	}
}

/* write outputs to server */
static void client_put_outputs(struct client_session *s)
{
	size_t len;
	struct buffer *ptr;

	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		if (s->outputs_sz && (ptr = par_image_outputs_buffer(s->pi))) {
			s->last_offset = 0;
			len = s->outputs_sz > PEER_MAX_MSG-6 ?
					PEER_MAX_MSG-6 : s->outputs_sz;
			debug("fd=%d Sending OUTPUTS_WRITE %lu at 0", s->sfd, len);
			buffer_put_u16(s->wb, 6 + len);
			buffer_put_u8(s->wb, CMD_OUTPUTS_WRITE);
			buffer_put_u8(s->wb, STATUS_OK);
			buffer_put_u16(s->wb, len);
			buffer_put_u16(s->wb, 0);
			/* TODO: this is not safe. */
			if (len <= buffer_size(ptr))
				buffer_put(s->wb, buffer_ptr(ptr), len);
			s->last_offset += len;
			s->state = CLIENT_OUTPUTS_WRITE;
			client_send(s);
			timeout_add(s->tout, s->response_timeout);
			debug("fd=%d CONNECTED->OUTPUTS_WRITE", s->sfd);
		} else {
			debug("fd=%d no outputs to write", s->sfd);
		}
	} else {
		warn("client_put_outputs(): sfd=%d", s->sfd);
	}
}

/* write inputs to server */
static void client_put_inputs(struct client_session *s)
{
	size_t len;
	struct buffer *ptr;

	if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
		if (s->inputs_sz && (ptr = par_image_inputs_buffer(s->pi))) {
			s->last_offset = 0;
			len = s->inputs_sz > PEER_MAX_MSG-6 ?
					PEER_MAX_MSG-6 : s->inputs_sz;
			debug("fd=%d Sending INPUTS_WRITE %lu at 0", s->sfd, len);
			buffer_put_u16(s->wb, 6 + len);
			buffer_put_u8(s->wb, CMD_INPUTS_WRITE);
			buffer_put_u8(s->wb, STATUS_OK);
			buffer_put_u16(s->wb, len);
			buffer_put_u16(s->wb, 0);
			if (len <= buffer_size(ptr))
				buffer_put(s->wb, buffer_ptr(ptr), len);
			s->last_offset += len;
			s->state = CLIENT_INPUTS_WRITE;
			client_send(s);
			timeout_add(s->tout, s->response_timeout);
			debug("fd=%d CONNECTED->INPUTS_WRITE", s->sfd);
		} else {
			debug("fd=%d no inputs to write", s->sfd);
		}
	} else {
		warn("client_put_inputs(): sfd=%d", s->sfd);
	}
}

/* status monitor stops io loop when all peers returned to connected state */
static void monitor_state()
{
	size_t in_progress = 0;
	slist_t iter;
	struct client_session *s;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		/* We monitor all sessions and count all
		   sessions which are still waiting for
		   response or timeout. */
		if (s->sfd >= 0 &&
		    (s->state != CLIENT_CONNECTED  &&
		     s->state != CLIENT_DISCONNECTED))
			in_progress++;
	}

	debug("monitor: %ld in progress", in_progress);

	/* Block on i/o only if at least one session still not done. */
	if (in_progress == 0)
		io_loop_stop(io);
}

static void response_recieve(int fd, struct buffer *pkt, struct client_session *s)
{
	u_int8_t cmd;
	u_int8_t status;
	u_int16_t plen;
	u_int16_t side;
	size_t len, offset;
	struct buffer *bp;

	if (buffer_size(pkt) < 4)
		return;

	plen = buffer_get_u16(pkt);
	cmd = buffer_get_u8(pkt);
	status = buffer_get_u8(pkt);

	debug("fd=%d CMD=%d status=%d", fd, cmd, status);

	switch (cmd) {
	case CMD_CONNECT:
		if (status == STATUS_OK && s->state == CLIENT_DISCONNECTED) {
			debug("fd=%d receive CMD_CONNECT: OK", fd);
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			s->state = CLIENT_CONNECTING;
			debug("fd=%d DISCONNECTED->CONNECTING", fd);
			client_obtain_config(s);
		} else if (status != STATUS_OK && s->state == CLIENT_DISCONNECTED) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d server refused connection", fd);
			s->status = status;
			s->state = CLIENT_DISCONNECTED;
		} else {
			debug("fd=%d unexpected CMD_CONNECT: %d", fd, status);
		}
		break;
	case CMD_OBTAIN_CONFIG:
		if (status == STATUS_OK && s->state == CLIENT_CONNECTING) {
			debug("fd=%d receive CMD_OBTAIN_CONFIG: OK", fd);
			if (buffer_size(pkt) >= 6) {
				if (timeout_active(s->tout))
					timeout_del(s->tout);
				s->peer_inputs_sz = buffer_get_u16(pkt);
				s->peer_outputs_sz = buffer_get_u16(pkt);
				side = buffer_get_u16(pkt);
				if ((s->side == PI_SIDE_MASTER && side != SIDE_MASTER) ||
				    (s->side == PI_SIDE_SLAVE && side != SIDE_SLAVE)) {
					warn("fd=%d server expected to be %s but he is not",
							s->sfd, s->side == PI_SIDE_MASTER ?
									"master" : "slave");
					s->state = CLIENT_DISCONNECTED;
					s->status = STATUS_SIDE_MISMATCH;
					debug("fd=%d CONNECTING->DISCONNECTED");
					return;
				}
				if ((bp = par_image_inputs_buffer(s->pi)) != NULL &&
				    s->peer_inputs_sz > 0) {
					s->inputs_sz = s->peer_inputs_sz < buffer_size(bp) ?
							s->peer_inputs_sz : buffer_size(bp);
				}
				if ((bp = par_image_outputs_buffer(s->pi)) != NULL &&
				    s->peer_outputs_sz > 0) {
					s->outputs_sz = s->peer_outputs_sz < buffer_size(bp) ?
							s->peer_outputs_sz : buffer_size(bp);
				}
				debug("Config: REMOTE I=%lu Q=%lu LOCAL I=%lu Q=%lu",
					s->peer_inputs_sz, s->peer_outputs_sz,
					s->inputs_sz, s->outputs_sz);
				s->state = CLIENT_CONNECTED;
				debug("fd=%d CONNECTING->CONNECTED", fd);
			}
			s->status = STATUS_OK;
		} else if (status != STATUS_OK && s->state == CLIENT_CONNECTING) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d could not obtain config CONNECTING->DISCONNECTED", fd);
			s->status = status;
			s->state = CLIENT_DISCONNECTED;
		} else {
			debug("fd=%d unexpected CMD_OBTAIN_CONFIG: %d", fd, status);
		}
		break;
	case CMD_PREPARE_READ:
		if (status == STATUS_OK && s->state == CLIENT_PREPARE_READ_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_PREPARE_READ: OK", fd);
			s->state = CLIENT_CONNECTED;
			s->status = STATUS_OK;
			debug("fd=%d PREPARE_READ_WAIT->CONNECTED", fd);
		} else if (status != STATUS_OK && s->state == CLIENT_PREPARE_READ_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_PREPARE_READ: %d", fd, status);
			s->status = status;
			s->state = CLIENT_CONNECTED;
			debug("fd=%d PREPARE_READ_WAIT->CONNECTED", fd);
		} else {
			debug("fd=%d unexpected CMD_PREPARE_READ: %d", fd, status);
		}
		break;
	case CMD_PREPARE_WRITE:
		if (status == STATUS_OK && s->state == CLIENT_PREPARE_WRITE_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_PREPARE_WRITE: OK", fd);
			s->state = CLIENT_CONNECTED;
			s->status = STATUS_OK;
			debug("fd=%d PREPARE_WRITE_WAIT->CONNECTED", fd);
		} else if (status != STATUS_OK && s->state == CLIENT_PREPARE_WRITE_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_PREPARE_WRITE: %d", fd, status);
			s->status = status;
			s->state = CLIENT_CONNECTED;
			debug("fd=%d PREPARE_WRITE_WAIT->CONNECTED", fd);
		} else {
			debug("fd=%d unexpected CMD_PREPARE_WRITE: %d", fd, status);
		}
		break;
	case CMD_INPUTS_READ:
		if (status == STATUS_OK && s->state == CLIENT_INPUTS_WAIT) {
			debug("fd=%d receive CMD_INPUTS_READ: OK", fd);
			/* put inputs to image */
			if (buffer_size(pkt) >= 4) {
				if (timeout_active(s->tout))
					timeout_del(s->tout);
				len = buffer_get_u16(pkt);
				offset = buffer_get_u16(pkt);
				if (buffer_size(pkt) == len &&
			            (bp = par_image_inputs_buffer(s->pi)) != NULL) {
					debug("fd=%d INPUTS len=%lu offset=%lu", fd, len, offset);
					if (offset < buffer_size(bp)) {
						len = offset + len > buffer_size(bp) ?
							buffer_size(bp) - offset : len;
						if (len)
							buffer_poke(bp, buffer_ptr(pkt), len, offset);
						debug("fd=%d wrote %lu bytes INPUTS", fd, len);
					}
				}
				if (s->last_offset < s->inputs_sz) {
					len = (s->inputs_sz - s->last_offset > PEER_MAX_MSG-6 ?
						PEER_MAX_MSG-6 : s->inputs_sz - s->last_offset);
					client_inputs_read(s, len, s->last_offset);
					s->last_offset += len;
				} else {
					s->state = CLIENT_CONNECTED;
					s->status = STATUS_OK;
					debug("fd=%d INPUTS_WAIT->CONNECTED", fd);
				}
			} else {
				if (timeout_active(s->tout))
					timeout_del(s->tout);
				debug("fd=%d bogus INPUTS_READ packet", fd);
				s->state = CLIENT_CONNECTED;
				s->status = STATUS_MALFORMED;
				debug("fd=%d INPUTS_WAIT->CONNECTED", fd);
			}
		} else if (status != STATUS_OK && s->state == CLIENT_INPUTS_WAIT) {
			debug("fd=%d receive CMD_INPUTS_READ: %d", fd, status);
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d INPUTS_WAIT->CONNECTED", fd);
			s->state = CLIENT_CONNECTED;
			s->status = status;
		} else {
			debug("fd=%d unexpected CMD_OBTAIN_CONFIG: %d", fd, status);
		}
		break;
	case CMD_OUTPUTS_READ:
		if (status == STATUS_OK && s->state == CLIENT_OUTPUTS_WAIT) {
			if (buffer_size(pkt) >= 4) {
				if (timeout_active(s->tout))
					timeout_del(s->tout);
				len = buffer_get_u16(pkt);
				offset = buffer_get_u16(pkt);
				if (buffer_size(pkt) == len &&
			            (bp = par_image_outputs_buffer(s->pi)) != NULL) {
					debug("fd=%d OUTPUTS len=%lu offset=%lu", fd, len, offset);
					if (offset < buffer_size(bp)) {
						len = offset + len > buffer_size(bp) ?
								buffer_size(bp) - offset : len;
						if (len)
							buffer_poke(bp, buffer_ptr(pkt), len, offset);
						debug("fd=%d wrote %lu bytes OUTPUTS", fd, len);
					}
				}
				if (s->last_offset < s->outputs_sz) {
					len = (s->outputs_sz - s->last_offset > PEER_MAX_MSG-6 ?
							PEER_MAX_MSG-6 : s->outputs_sz - s->last_offset);
					client_outputs_read(s, len, s->last_offset);
					s->last_offset += len;
				} else {
					s->state = CLIENT_CONNECTED;
					s->status = STATUS_OK;
					debug("fd=%d OUTPUTS_WAIT->CONNECTED", fd);
				}
			} else {
				if (timeout_active(s->tout))
					timeout_del(s->tout);
				debug("fd=%d bogus OUTPUTS_READ packet", fd);
				s->state = CLIENT_CONNECTED;
				s->status = STATUS_MALFORMED;
				debug("fd=%d OUTPUTS_WAIT->CONNECTED", fd);
			}
		} else if (status != STATUS_OK && s->state == CLIENT_OUTPUTS_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_OUTPUTS_READ : %d", fd, status);
			s->state = CLIENT_CONNECTED;
			s->status = status;
		} else {
			debug("fd=%d unexpected CMD_OUTPUTS_READ: %d", fd, status);
		}
		break;
	case CMD_OUTPUTS_WRITE:
		if (status == STATUS_OK && s->state == CLIENT_OUTPUTS_WRITE) {
			debug("fd=%d receive CMD_OUTPUTS_WRITE: OK", fd);
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			if (s->last_offset < s->outputs_sz &&
			    (bp = par_image_outputs_buffer(s->pi))) {
				len = s->outputs_sz - s->last_offset > PEER_MAX_MSG-6 ?
					PEER_MAX_MSG-6 : s->outputs_sz - s->last_offset;
				debug("fd=%d Sending OUTPUTS_WRITE %lu at %lu", fd, len, s->last_offset);
				buffer_put_u16(s->wb, 6 + len);
				buffer_put_u8(s->wb, CMD_OUTPUTS_WRITE);
				buffer_put_u8(s->wb, STATUS_OK);
				buffer_put_u16(s->wb, len);
				buffer_put_u16(s->wb, s->last_offset);
				/* TODO: this is not safe. */
				if (len + s->last_offset <= buffer_size(bp))
					buffer_put(s->wb, buffer_ptr(bp) + s->last_offset, len);
				s->last_offset += len;
				client_send(s);
				timeout_add(s->tout, s->response_timeout);
			} else {
				s->state = CLIENT_CONNECTED;
				s->status = STATUS_OK;
				debug("fd=%d OUTPUTS_WRITE->CONNECTED", fd);
			}
		} else if (status != STATUS_OK && s->state == CLIENT_OUTPUTS_WRITE) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_OUTPUTS_WRITE: %d", fd, status);
			s->status = status;
			s->state = CLIENT_CONNECTED;
			debug("fd=%d OUTPUTS_WRITE->CONNECTED", fd);
		} else {
			debug("fd=%d unexpected CMD_OUTPUTS_READ: %d", fd, status);
		}
		break;
	case CMD_INPUTS_WRITE:
		if (status == STATUS_OK && s->state == CLIENT_INPUTS_WRITE) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_INPUTS_WRITE: OK", fd);
			if (s->last_offset < s->inputs_sz &&
			    (bp = par_image_inputs_buffer(s->pi))) {
				len = s->inputs_sz - s->last_offset > PEER_MAX_MSG-6 ?
						PEER_MAX_MSG-6 : s->inputs_sz - s->last_offset;
				debug("fd=%d Sending INPUTS_WRITE %lu at %lu", fd, len, s->last_offset);
				buffer_put_u16(s->wb, 6 + len);
				buffer_put_u8(s->wb, CMD_INPUTS_WRITE);
				buffer_put_u8(s->wb, STATUS_OK);
				buffer_put_u16(s->wb, len);
				buffer_put_u16(s->wb, s->last_offset);
				if (len + s->last_offset <= buffer_size(bp))
					buffer_put(s->wb,
						buffer_ptr(bp) + s->last_offset, len);
				s->last_offset += len;
				client_send(s);
				timeout_add(s->tout, s->response_timeout);
			} else {
				s->state = CLIENT_CONNECTED;
				s->status = STATUS_OK;
				debug("fd=%d INPUTS_WRITE->CONNECTED", fd);
			}
		} else if (status != STATUS_OK && s->state == CLIENT_INPUTS_WRITE) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			debug("fd=%d receive CMD_INPUTS_WRITE: %d", fd, status);
			s->status = status;
			s->state = CLIENT_CONNECTED;
			debug("fd=%d INPUTS_WRITE->CONNECTED", fd);
		} else {
			debug("fd=%d unexpected CMD_INPUTS_WRITE: %d", fd, status);
		}
		break;
	case CMD_END_READ:
		if (status == STATUS_OK && s->state == CLIENT_END_READ_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			else
				warn("fd=%d timeout MUST run whem in END_READ_WAIT state!", fd);
			debug("fd=%d receive CMD_END_READ: OK", fd);
			s->state = CLIENT_CONNECTED;
			s->status = STATUS_OK;
			debug("fd=%d END_READ_WAIT->CONNECTED", fd);
		} else if (status != STATUS_OK && s->state == CLIENT_END_READ_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			else
				warn("fd=%d timeout MUST run when in END_READ_WAIT state!", fd);
			debug("fd=%d receive CMD_END_READ: %d", fd, status);
			s->state = CLIENT_CONNECTED;
			s->status = status;
			debug("fd=%d END_READ_WAIT->CONNECTED", fd);
		} else {
			debug("fd=%d unexpected CMD_END_READ: %d", fd, status);
		}
		break;
	case CMD_END_WRITE:
		if (status == STATUS_OK && s->state == CLIENT_END_WRITE_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			else
				warn("fd=%d timeout MUST run when END_WRITE_WAIT state!", fd);
			s->state = CLIENT_CONNECTED;
			s->status = STATUS_OK;
			debug("fd=%d receive CMD_END_WRITE: OK", fd);
		} else if (status != STATUS_OK && s->state == CLIENT_END_WRITE_WAIT) {
			if (timeout_active(s->tout))
				timeout_del(s->tout);
			else
				warn("fd=%d timeout MUST run whem in END_WRITE_WAIT state!", fd);
			s->state = CLIENT_CONNECTED;
			s->status = status;
		} else {
			debug("fd=%d unexpected CMD_END_WRITE: %d", fd, status);
		}
		break;
	default:
		/* only warn user on bad command don't change state */
		warn("pi_client: fd=%d unknown command %d", fd, cmd);
		s->status = STATUS_BAD_COMMAND;
		break;
	}
}

static void client_event(int fd, short events, void *opaque)
{
	int res;
	int state_before;
	u_int16_t len;
	unsigned char pbuf[PEER_MAX_MSG+2];
	buffer_t pkt;
	struct client_session *s = (struct client_session *) opaque;

	if (events & EVENT_ERROR) {
		debug("fd=%d error", s->sfd);
		client_session_disable(s, STATUS_NOT_AVAIL);
		monitor_state();
		return;
	}

	if (events & EVENT_TIMEOUT) {
		/* server couldn't respond in time */
		debug("fd=%d timeout", s->sfd);
		client_session_disable(s, STATUS_TIMEOUT);
		monitor_state();
		return;
	}

	if (events & EVENT_READ) {
		res = buffer_read_fd(s->sfd, s->rb, 4096);
		if (res == 0) {
			/* server closed connection */
			client_session_disable(s, STATUS_NOT_AVAIL);
			monitor_state();
			return;
		} else  if (res == -1 &&
		    (errno != EINTR && errno != EAGAIN &&
				    errno != EWOULDBLOCK)) {
			warn("client_event: read()=%d sock=%d: %s",
					res, s->sfd, strerror(errno));
		} else if (res > 0) {
			state_before = s->state;
			pkt = buffer_create_static(pbuf, sizeof(pbuf));
			while (buffer_size(s->rb) >= 4) {
				len = buffer_get_u16_at(s->rb, 0);
				if (buffer_size(s->rb) >= len + 2 &&
				    len <= PEER_MAX_MSG) {
					buffer_copy(pkt, s->rb, len + 2);
					buffer_consume(s->rb, len + 2);
					response_recieve(fd, pkt, s);
					buffer_reset(pkt);
				} else if (len > PEER_MAX_MSG) {
					buffer_reset(s->rb);
					buffer_reset(pkt);
				}
					break;
			}
			/* Run monitor if session state changed to non-waiting. */
			if (state_before != s->state &&
			    (s->state == CLIENT_CONNECTED ||
			     s->state == CLIENT_DISCONNECTED))
				monitor_state();
			buffer_free(pkt);
		}
	}

	if (events & EVENT_WRITE) {
		if (buffer_size(s->wb) != 0) {
			res = buffer_write_fd(s->sfd, s->wb);
			if (res == -1 && (errno != EINTR &&
				errno != EWOULDBLOCK && errno != EAGAIN)) {
				warn("client_event: write()=%d sock=%d: %s",
						res, s->sfd, strerror(errno));
			}
		}
		if (buffer_size(s->wb) == 0)
			io_change(io, s->sfd, EVENT_READ);
	}
}

struct pclt *pclt_new(const char *sock, int timeout, size_t inputs, size_t outputs, pi_mode_t mode, pi_side_t side)
{
	int sfd;
	struct client_session *s = NULL;
	slist_t list;

	if (!sock)
		error("pi_add: sock NULL");

	sfd = sock_unix_connect(sock);
	if (sfd == -1)
		return NULL;
	sock_nonblock(sfd);
	s = xmalloc(sizeof(*s));
	s->sfd = sfd;
	s->rb = buffer_new();
	s->wb = buffer_new();
	xstrncpy(s->sockfile, sock, sizeof(s->sockfile));
	s->tout = timeout_new(client_event, s);
	s->state = CLIENT_DISCONNECTED;
	s->last_offset = 0;
	s->side = side;
	s->pi = par_image_new(inputs, outputs, mode);
	s->peer_inputs_sz = 0;
	s->peer_outputs_sz = 0;
	s->inputs_sz = 0;
	s->outputs_sz = 0;
	if (timeout > 0)
		s->response_timeout = timeout;
	else
		s->response_timeout = PEER_TIMEOUT_MSEC;
	list = slist_new();
	list->data = s;
	client_list = slist_add(client_list, list);
	io_add(io, sfd, EVENT_READ, client_event, s);
	return (struct pclt *) s;
}

int pclt_add_par(struct pclt *c, struct par_struct *par, par_dir_t dir, unsigned int offset, unsigned char bit)
{
	struct client_session *s = (struct client_session *) c;

	return par_image_add_par(s->pi, par, dir, offset, bit);
}

static void client_session_free(struct client_session *s)
{
	slist_t list;

	if (s->sfd >= 0) {
		io_del(io, s->sfd);
		close(s->sfd);
	}
	if (timeout_active(s->tout))
		timeout_del(s->tout);
	buffer_free(s->wb);
	buffer_free(s->rb);
	timeout_free(&s->tout);
	list = slist_find(client_list, s, NULL);
	if (list != NULL)
		client_list = slist_unlink(client_list, list);
	slist_free(&list);
	par_image_free(&s->pi);
	xfree(s);
}

static void client_session_disable(struct client_session *s, int status)
{
	if (s->sfd < 0)
		warn("session %p already disabled", s);
	debug("disabling session: fd=%d cause=%d", s->sfd, status);
	if (timeout_active(s->tout))
		timeout_del(s->tout);
	buffer_reset(s->rb);
	buffer_reset(s->wb);
	if (s->sfd >= 0) {
		io_del(io, s->sfd);
		close(s->sfd);
	}
	s->peer_inputs_sz = 0;
	s->peer_outputs_sz = 0;
	s->inputs_sz = 0;
	s->outputs_sz = 0;
	s->state = CLIENT_DISCONNECTED;
	s->status = status;
	s->last_offset = 0;
	s->sfd = -1;
}

void pclt_connect()
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->state == CLIENT_DISCONNECTED) {
			client_connect(s);
			if (s->state != CLIENT_CONNECTED)
				in_progress++;
		}
	}

	if (in_progress)
		io_loop_run(io);
}

void pclt_try_reconnect()
{
	int sfd;
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->sfd < 0 && s->state == CLIENT_DISCONNECTED) {
			sfd = sock_unix_connect(s->sockfile);
			if (sfd == -1)
				continue;
			sock_nonblock(sfd);
			s->sfd = sfd;
			io_add(io, sfd, EVENT_READ, client_event, s);
			client_connect(s);
			if (s->state != CLIENT_CONNECTED)
				in_progress++;
		}
	}

	if (in_progress)
		io_loop_run(io);
}

/* Ask server to prepare inputs  */
void pclt_prepare_inputs(pi_side_t side)
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == side) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				if (side == PI_SIDE_SLAVE)
					client_prepare_inputs(s);
				else if (side == PI_SIDE_MASTER) {
					par_image_update_inputs(s->pi, PI_SIDE_MASTER);
					client_prepare_outputs(s);
				}
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);
}

/* Ask server to prepare outputs */
void pclt_prepare_outputs(pi_side_t side)
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		/* update point values to or from image depends on side */
		if (s->side == side) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				if (side == PI_SIDE_SLAVE) {
					par_image_update_outputs(s->pi, PI_SIDE_SLAVE);
					client_prepare_outputs(s);
				} else if (side == PI_SIDE_MASTER)
					client_prepare_inputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);
}

/* read slave side inputs - we need current input and output values */
void pclt_read_inputs()
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_SLAVE) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_get_inputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

	in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_SLAVE) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_get_outputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

	in_progress = 0;

	/* Notify all connected slaves that we are done with reading and update points from image */
	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_SLAVE && s->state == CLIENT_CONNECTED) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				par_image_update_inputs(s->pi, PI_SIDE_SLAVE);
				client_end_read(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);
}

/* write master side inputs - master wants to read current input and output values */
void pclt_write_inputs()
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_MASTER) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_put_inputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

	in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_MASTER) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_put_outputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

	in_progress = 0;

	/* Notify all connected masters that we are done with writing */
	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_MASTER) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_end_write(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);
}

/* read master outputs - we will write them */
void pclt_read_outputs()
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_MASTER) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_get_outputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

	in_progress = 0;

	/* Notify all connected masters that we are done with reading */
	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_MASTER) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				par_image_update_outputs(s->pi, PI_SIDE_MASTER);
				client_end_read(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);
}

/* write slave outputs */
void pclt_write_outputs()
{
	slist_t iter;
	struct client_session *s;
	size_t in_progress = 0;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_SLAVE) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_put_outputs(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

	in_progress = 0;

	/* Notify all connected slaves that we are done with writing */
	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s->side == PI_SIDE_SLAVE) {
			if (s->sfd >= 0 && s->state == CLIENT_CONNECTED) {
				client_end_write(s);
				if (s->state != CLIENT_CONNECTED)
					in_progress++;
			}
		}
	}

	if (in_progress)
		io_loop_run(io);

}

void pclt_enumerate(pclt_enum_callback_t callback)
{
	int res;
	slist_t iter;
	struct client_session *s;

	if (!callback)
		return;

	for (iter = client_list; iter != NULL; iter = iter->next) {
		s = (struct client_session *) iter->data;
		if (s) {
			res = (*callback)(s->sockfile, s->state, s->status);
			if (res != 0)
				break;
		} else {
			warn("session %p has no data", iter);
		}
	}
}

void pclt_init()
{
	if (io != NULL) {
		warn("pi_init: already initialized");
		return;
	}

	info("initializing PI client");

	io = io_new();
}

void pclt_deinit()
{
	slist_t iter, next;
	struct client_session *s;

	info("deinitializing PI client");

	iter = client_list;
	while (iter) {
		next = iter->next;
		s = (struct client_session *) iter->data;
		client_session_free(s);
		iter = next;
	}

	if (io)
		io_free(&io);
}
